module cpu(rst, clk, irq,
	   pc, inst, sjnzp,
	   flu, bubble, indirect, indirected, stall,
	   ra1, rd1, ra2, rd2,
	   we1, wa1, wd1, we2, wa2, wd2,
   	   apsr, memw, mem_out,
	   push_stack, pop_stack, savepc_to_R7,
	   in_irq, in_exc, rti);

   input rst;	// low for reset, high for work
   input clk;	// posedge for logic and calc, negedge for memory storage
   input irq;	// high for interrupt request

   output [15:0] inst;		// the current instruction in FETCH stage

   output	pc;
   reg [15:0]	pc;		// program counter
   wire [15:0]	s2_pc;		// the new pc if jump

   // stage registers
   wire [65:0]	s0;		// stage FETCH (not reg)
   reg [65:0]	s1;		// stage between FETCH and EXEC
   reg [65:0]	s2;		// stage between EXEC and MEM

   // jump
   output [4:0] sjnzp;		// save save PC => R7, unconditional jump, jump if negative, zero, positive
   wire 	nzp_match;
   assign	nzp_match = (apsr[2:0] & sjnzp[2:0]) != 3'b000;
   
   // judge whether to stall by the current ops (see stall.v)
   wire 	trap;		// this is a trap instruction
   wire		load_mem;	// should read from memory
   wire 	arith_alu;	// R-type inst in EXEC stage
   wire		indirect_mem;	// need indirect mem 

   // altering the normal execution flow
   output	flu;		// flush pipeline
   assign	flu = nzp_match | sjnzp[3] | trap |
			entering_irq |
			entering_exc |
			entering_rti;
   output	bubble;		// pause PC, ALU, register, clear memory 
   output	indirect;	// pause PC, pipeline, register and memory write
   output	indirected;	// the previous stage was indirected
   output	stall;		// pause pipeline
   assign	stall = indirect | bubble;
   
   // irq and exception special subroutines
   wire		irq_stage1; 	// PC => M[R6], R6--
   wire		irq_stage2;	// PSR => M[R6], R6--
   wire		irq_stage3;	// read irq addr: wd = M[0x40]
   wire		irq_done;	// PC <= wd (this pin is unused)
   output	in_irq;		// we are in irq handler
   wire		entering_irq;
   
   wire		exc;		// exception inst: PC => M[R6], R6--
   wire		exc_stage1;	// PSR => M[R6], R6--
   wire		exc_stage2;	// read exception addr: wd = M[0x44]
   wire		exc_done;	// PC <= wd (this pin is unused)
   output	in_exc;		// we are in exception handler
   wire		entering_exc;
   
   output	rti;		// RTI inst: R6++; PSR <= M[R6]
   wire		rti_stage1;	// RTI inst: R6++; PC <= M[R6]
   wire		entering_rti;

   // Read/Write registers
   // read port 1 and 2 are both in EXEC stage, linked to ALU's input
   output [2:0]  ra1, ra2;	// read address
   output [15:0] rd1, rd2;	// read data
   // write port 1 is in EXEC stage, linked to ALU's output
   // write port 2 is in MEM stage, linked to MEM's output
   output        we1, we2;	// write enable
   output [2:0]  wa1, wa2;	// write address
   output [15:0] wd1, wd2;	// write data

   // Arithmetic Program State Register
   output [15:0] apsr;		// only the low 5 bits are calculated
   wire [4:0] 	 apsr_mem;	// apsr from MEM and ALU are calculated separately
   wire [4:0] 	 apsr_alu;
   wire		 apsr_is_mem;	// apsr should come from MEM
   wire 	 apsr_is_alu;	// apsr should come from ALU

   // Read/Write memory
   output 	 memw;		// memory write enable
   reg [15:0] 	 store_data;	// data to be written
   reg [15:0] 	 alu_out;	// (possibly) write address
   output [15:0] mem_out;	// data read from memory
   assign wd2 = mem_out;	// memory out is sent to register if we2
   
   // stack operation
   output push_stack;
   assign push_stack = irq_stage1 | irq_stage2 | exc | exc_stage1;
   output pop_stack;
   assign pop_stack = rti | rti_stage1;
   wire   stack_pc;		// push/pop PC
   assign stack_pc = irq_stage1 | exc | rti;
   wire   stack_apsr;		// push/pop PSR
   assign stack_apsr = irq_stage2 | exc_stage1 | rti_stage1;
   wire [15:0] s2_imm;		// immediate from stage 2
   
   // Generic Registers
   output savepc_to_R7;		// PC needs to be saved to R7
   assign savepc_to_R7 = sjnzp[4];

   greg greg_i(rst, clk, flu,
	       ra1, rd1,
	       // if loadpc_from_stack, PC will come out from rd2
	       ra2, rd2,
	       // if bubble or indirect, ALU stage should not do any write
	       we1 & ~stall, wa1, wd1,
	       // if bubble, memory stage should execute and write back
	       // if indirect, memory stage should not write
	       we2 & ~indirect, wa2, wd2,
	       // Save whose PC? From MEM stage, not the current PC!
	       // if savepc happen when pipeline is flushed, then use pc instead
	       s2_pc == 0 ? pc : s2_pc,
	       // Save PC if savepc is set for MEM stage inst
	       savepc_to_R7,
	       push_stack, pop_stack);

   // register for Arithmetic Program Status Register
   sreg sreg_apsr(.rst(rst),
   		  .clk(clk),
		  .flu(flu),
		  .rd1(apsr),
		  .we1(apsr_is_alu),
		  .wd1({11'b0, apsr_alu}),
		  .we2(apsr_is_mem),
		  .wd2({11'b0, apsr_mem}),
		  .wd_flu(rti), // load PSR from memory
		  .wdf(mem_out));
  
   // stage FETCH
   // instruction memory
   instmem instmem_i(.subclk(~clk),
	   	     .pc(pc),
		     .ir(inst));

   decoder decoder_i(.pc(pc),
	   	     .ir(inst),
		     .state(s0));
   
   // stage EXEC
   state_unpack su_i1(.state(s1),
		      .sr1a(ra1),
		      .sr2a(ra2),
		      .dra(wa1),
		      .iswb_alu(we1),
		      .apsr_is_alu(apsr_is_alu),
		      .arith_alu(arith_alu));

   stage_exec se_i(.state(s1),
	   	   .rd1(rd1),
		   .rd2(rd2),
		   .wd(wd1),
		   .apsr(apsr_alu));

   // stage MEM
   state_unpack su_i2(.state(s2),
		      .imm(s2_imm),
		      .pc(s2_pc),
		      .iswb_mem(we2),
		      .dra(wa2),
		      .sjnzp(sjnzp),
		      .memw(memw),
		      .apsr_is_mem(apsr_is_mem),
		      .load_mem(load_mem),
		      .indirect_mem(indirect_mem),
		      .trap(trap),
		      .exc(exc),
		      .rti(rti));
   
   stage_mem sm_i(.clk(clk),
		  // if stack operation, then the memory address should be R6
		  .addr((pop_stack | push_stack) ? rd2 : 
			  (irq_stage3 ? 8'h40 : // if interrupt stage 2, read 0x40 for interrupt handler addr
			   (exc_stage2 ? 8'h44 : // exception handler at M[0x44]
			    (trap ? s2_imm : alu_out) // goto trap vector if TRAP; alu_out in normal case
			   )
			  )
			 ),
		  .rd(mem_out),
		  // if push_stack, then write data and write enable should be handled specially
		  .wd(push_stack ?
			(stack_apsr ? apsr : (s2_pc == 0 ? pc : s2_pc))
			: store_data),
		  .we(push_stack ? 1'b1 : memw),
		  .apsr(apsr_mem),
		  .indirect_stage(indirected),
		  .indirect_inst(indirect_mem));

   // control for pausing pipeline
   stall stall_i(rst, clk,
		 flu, bubble, indirect, indirected,
		 load_mem, arith_alu, indirect_mem,
		 ra1, ra2,
		 wa1, wa2);

   irq irq_i(rst, clk, irq, in_exc, rti, in_irq,
		irq_stage1, irq_stage2, irq_stage3, irq_done, entering_irq);
   exc exc_i(rst, clk, exc, rti, in_exc,
		exc_stage1, exc_stage2, exc_done, entering_exc);
   rti rti_i(rst, clk, rti,
		rti_stage1, entering_rti);
   
   // flow control (PC, s1, s2)
   always@(negedge rst or posedge clk)
     if(~rst)
       begin
	  pc <= 8'h60; // The starting address is 0x60
	  s1 <= 0;
	  s2 <= 0;
       end
     else if(irq)
       // Because irq is async, the irq cycle do nothing but preventing write.
       // The handling of irq starts next cycle (irq_stage1).
       // If flush, then savepc need current pc after jump
       // Else, cannot clear s2 because savepc need pc from s2
       begin
	  s1 <= 0;
	  if (flu)
		s2 <= 0;
       end
     else if(flu)
       begin
       	  s1 <= 0;
	  s2 <= 0;
	  // save pc, psr to stack; or pop psr from stack
	  if (irq_stage1 | irq_stage2 | exc | exc_stage1 | rti)
		;
          // jump to interrupt/trap/exception vector or stack addr
	  else if (irq_stage3 | exc_stage2 | trap | rti_stage1)
		pc <= mem_out;
	  // otherwise the result of ALU is PC + offset or Register n
	  else
	  	pc <= alu_out;
       end
     else if(indirect)
       // in indirect stage, MEM stage need to be replayed
       // EXEC stage should do nothing
       ;
     else if(bubble)
       // EXEC and MEM stage should do nothing
       begin
	  s2 <= 0;
       end
     else
       // The normal case:
       // PC should increment;
       // state should flow to the next;
       // if memw, data from EXEC stage is to be stored to MEM in the next stage,
       //    where the data is read from register,
       //    and the address is the output of ALU.
       begin
	  pc <= pc+1;
	  s1 <= s0;
	  s2 <= s1;

	  store_data <= rd2;
	  alu_out <= wd1;
       end

endmodule // cpu
