module sreg(rst, clk, flu, rd1, rd2, we1, wd1, we2, wd2, wd_flu, wdf);
/**
 * Every register are internally two registers d1 and d2.
 * Register read is asynchronous.
 * d1 updates its value from EXEC stage at posedge clk.
 * d2 updates its value at posedge clk:
 *   - from d1
 *   - from MEM stage
 * Hence when an instruction whose output is from ALU
 *   flows from EXEC to MEM, the result is written when MEM stage starts,
 *   enabling the following instruction to get the updated value async.
 * If it is about to flush:
 *   - if a special value needs to be stored (e.g. save PC => R7),
 *     then d1 and d2 should be overwitten by wdf
 *   - otherwise d1 is overwitten by d2
 * If EXEC stage needs to write register:
 *   - if MEM stage also needs to write the same register, since
 *     the instruction at EXEC stage is newer, d1 should follow EXEC stage,
 *     d2 should follow MEM stage, and d1 will be written into d2 next cycle.
 *   - otherwise wd1 => d1
 * If MEM stage needs to write register:
 *   - if EXEC stage also needs to write the same register, as above.
 *   - otherwise wd2 from MEM stage is the latest value of the register,
 *     in case d1 overwrites it in the next cycle, d1 should be also written
 *     as wd2.
 */
   input rst, clk, flu;
   input we1, we2, wd_flu;
   input [15:0] wd1, wd2, wdf;
   output [15:0] rd1, rd2;

   reg [15:0] 	 d1, d2;

   assign rd1 = d1;
   assign rd2 = d2;
   
   always@(negedge rst or posedge clk)
     if(~rst)
       begin
	  d1 <= 0;
	  d2 <= 0;
       end	  
     else if(flu)
       if(wd_flu)
	 begin
	    d1 <= wdf;
	    d2 <= wdf;
	 end
       else
	 begin
	    d1 <= d2;
	 end
     else if(we1)
       if(we2)
	 begin
	    d1 <= wd1;
	    d2 <= wd2;
	 end
       else
	 begin
	    d1 <= wd1;
	    d2 <= d1;
	 end
     else if(we2)
       begin
	  d1 <= wd2;
	  d2 <= wd2;
       end
     else
       begin
	  d2 <= d1;
       end
endmodule // sreg
